
PRIMITIVE_TYPES = {
    bool,
    str,
    list,
    dict,
    float,
    int,
}

ACCEPTABLE_KEY_TYPES = {
    str,
    int,
    float
}

TYPED_TYPES_TO_PRIMITIVES = {
    'Dict': dict,
    'List': list,
    'Set': set,
}
